# NATools - Package with parallel corpora tools
# Copyright (C) 2002-2005  Alberto Simes
#
# This package is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 02111-1307, USA.

package NAT::Matrix;

use 5.006;
use strict;
use warnings;
use CGI qw/:standard/;
require Exporter;

use NAT::Client;
use Lingua::PT::PLNbase;
use Data::Dumper;
use IO::Socket;

our $VERSION = '0.02';

=head1 NAME

NAT::Matrix - Module to align a sentence by words

=head1 SYNOPSIS

  use NAT::Matrix;

=head1 DESCRIPTION

=cut

sub new {
  my ($class, $crpId, $s1, $s2) = @_;

  my $self = bless { sentence => [undef, $s1,$s2] }, $class;
  $self->{crp} = $crpId;
  $self->{s1} = [atomiza(lc($s1))];
  $self->{s2} = [atomiza(lc($s2))];

  # Using scalar we get the number of elements... not the last index
  $self->{size} = [undef,
		   scalar(@{$self->{s1}}),
		   scalar(@{$self->{s2}})];

  my ($i,$j);
  $i = 0;
  for $s1 (@{$self->{s1}}) {
    $j=0;
    for $s2 (@{$self->{s2}}) {
      $self->{matrix}[$i][$j] = $self->prob($s1,$s2);
      $j++;
    }
    $i++;
  }

  return $self;
}

sub findDiagonal {
  my ($self) = @_;

  $self->initResultMatrix;

  $self->setProbabilitiesForEqualStrings;

  $self->enhanceDiagonal;

  $self->markPoints;

  $self->markPatterns;

  my ($x,$y) = (0,0);

  $self->connectD($x,$y);

  return $self;

}

sub setProbabilitiesForEqualStrings {
  my ($self) = @_;

  for my $i (0..$self->{size}[1]-1) {
    next if $self->{s1}[$i] =~ m/^[aeiou]$/;
    for my $j (0..$self->{size}[2]-1) {
      next if $self->{s2}[$j] =~ m/^[aeiou]$/;

      if ($self->{s1}[$i] eq $self->{s2}[$j]) {
	$self->{matrix}[$i][$j] = 80
      }
    }
  }
}

sub initResultMatrix {
  my ($self) = @_;
  for my $i (0..$self->{size}[1]-1) {
    for my $j (0..$self->{size}[2]-1) {
      $self->{res}[$i][$j] = 0;
    }
  }
}



sub grep_blocks {
  my ($self) = @_;

  my $blocks;

  my ($i,$j) = (0,0);

  while ($i < $self->{size}[1]) {
    $j = 0;
  LOOP: while ($j < $self->{size}[2]) {
      my ($ii, $jj) = ($i,$j);
      while ($ii<$self->{size}[1] && $self->{res}[$ii][$j]) { $ii++ };
      while ($jj<$self->{size}[2] && $self->{res}[$i][$jj]) { $jj++ };

      my ($h, $w) = ($ii-$i, $jj-$j);
      if ($h && $w) {
	push @$blocks, { i=> $i, j=> $j, h=> $h, w=> $w };

	if ($self->{res}[$ii-1][$jj-1] eq "X") {
	  $ii--;
	  $jj--;
	  for (;;) {
	    if ($self->{res}[$ii-1][$jj] eq "X") {
	      $ii--
	    } elsif ($self->{res}[$ii][$jj-1] eq "X") {
	      $jj--
	    } else {
	      last
	    }
	  }
	  $i = $ii;
	  $j = $jj;
	} else {
	  if (($j+$w<$self->{size}[2] && $self->{res}[$i+$h-1][$j+$w]) ||
	      ($i+$h<$self->{size}[1] && $self->{res}[$i+$h][$j+$w-1])) {
	    $i += $h - 1;
	    $j += $w - 1;
	  } else {
	    $i += $h ;
	    $j += $w ;
	  }
	}
	next LOOP;
      }
      $j++;
    }
    $i++;
  }
  return $blocks
}

sub combine_blocks {
  my ($self, $blocks, $size) = @_;

  my $nr_blocks = scalar(@$blocks);

  my $combined_blocks;
  for (0..$nr_blocks-$size) {
    my $block = { i => $blocks->[$_]{i}, j => $blocks->[$_]{j} };
    $block->{h} = $blocks->[$_+$size-1]{i} + $blocks->[$_+$size-1]{h} -  $block->{i};
    $block->{w} = $blocks->[$_+$size-1]{j} + $blocks->[$_+$size-1]{w} -  $block->{j};
    push @$combined_blocks, $block;
  }
  return $combined_blocks;
}

sub dump_block {
  my ($self,$b) = @_;

  return [join(" ", map { $self->{s1}[$_] } ($b->{i}..$b->{i}-1+$b->{h})),
	  join(" ", map { $self->{s2}[$_] } ($b->{j}..$b->{j}-1+$b->{w}))];
}

sub relate {
  my ($self,$blocks) = @_;

  my $ans;

  for my $b (@$blocks) {
    push @$ans, [join(" ", map { $self->{s1}[$_] } ($b->{i}..$b->{i}-1+$b->{h})),
		 join(" ", map { $self->{s2}[$_] } ($b->{j}..$b->{j}-1+$b->{w}))]
  }

  return $ans;
}


sub connectD {
  my ($self,$i,$j) = @_;

  return $self if $i >= $self->{size}[1] || $j >= $self->{size}[2];

  return $self->connectD($i  ,$j+1) if ($self->{res}[$j+1][$i] eq 'P');
  return $self->connectD($i+1,$j  ) if ($self->{res}[$j][$i+1] eq 'P');
  return $self->connectD($i+1,$j+1) if ($self->{res}[$j+1][$i+1] eq 'P');

  my $x = 1;
  my ($pi,$pj) = ($i+1,$j+1);
  my ($nextstarti, $nextstartj) = (undef, undef);
  my ($starti, $startj) = ($i,$j);

 SEARCH: while($x+$i < max($self->{size}[1],$self->{size}[2]) &&
	       $x+$j < max($self->{size}[1],$self->{size}[2])) {
    if ($self->{res}[$j][$i] eq "X") {
      for (;;) {
	if ($self->{res}[$j+1][$i] eq 'X') {
	  $j++
	} elsif ( $self->{res}[$j][$i+1] eq 'X') {
	  $i++
	} else {
	  last
	}
      }
    }

    for my $ii ($i..$i+$x) {
      for my $jj ($j..$j+$x) {
	next if $i == $ii && $j == $jj;
	if ($self->{res}[$jj][$ii] eq 'P' || $self->{res}[$jj][$ii] eq 'X') {

	  if ($self->{res}[$jj][$ii] eq 'X') {
	    ($nextstarti, $nextstartj) = ($ii,$jj);
	    for (;;) {
	      if ($self->{res}[$jj+1][$ii] eq 'X') {
		$jj++
	      } elsif ( $self->{res}[$jj][$ii+1] eq 'X') {
		$ii++
	      } else {
		last
	      }
	    }
	  }

	  $pi = $ii;
	  $pj = $jj;
	  last SEARCH;
	}
      }
    }
    $x++;
  }
  for my $ii ($starti..$pi) {
    for my $jj ($startj..$pj) {
      $self->{res}[$jj][$ii] = "G" unless $self->{res}[$jj][$ii] eq "X"
    }
  }
  $i = $nextstarti || $pi;
  $j = $nextstartj || $pj;

  $self->connectD($i,$j);

  return $self;
}


=head2 C<max>

This function...

=cut

sub max {
  $_[0]>$_[1]?$_[0]:$_[1]
}

=head2 C<markPatterns>

This function...

=cut

sub markPatterns {
  my $self = shift;

  my @patterns = (
		  [[0,0,"P"],[0,"P",0],["P",0,0]],
		  [[0,"P"],["P",0]],
		  [[0,0,"P"],["P",0,0],[0,"P",0]],
		  [[0,0,"P"],["P",0,0]],
		 );

  for my $pat (@patterns) {
    $self->markPattern($pat);
  }

  return $self;
}

sub markPattern {
  my ($self,$P) = @_;
  my ($Px,$Py) = (scalar @$P, scalar @{$P->[0]});

  for my $x (0..$self->{size}[1]-$Px) {
    for my $y (0..$self->{size}[2]-$Py) {

      my $found = 1;
      for my $XX ($x..$x+$Px-1) {
	for my $YY ($y..$y+$Py-1) {
	  $found = 0 unless $self->{res}[$XX][$YY] eq $P->[$XX-$x][$YY-$y]
	}
      }

      if ($found) {
	push @{$self->{patterns}}, { i=>$x, j=>$y, h=>$Px, w=>$Py};
	for my $XX ($x..$x+$Px-1) {
	  for my $YY ($y..$y+$Py-1) {
	    $self->{res}[$XX][$YY] = "X"
	  }
	}
      }
    }
  }

  return $self;
}

=head2 C<markPoints>

This function...

=cut

sub markPoints {
  my $self = shift;

  my $p = 4 * 2 / ($self->{size}[1]+$self->{size}[2]);

  $p = $p > .2 ? $p : .2;

  for my $i (0..$self->{size}[1]-1) {
    for my $j (0..$self->{size}[2]-1) {

      my $maxR = $self->getMaxRow20($i);
      my $maxC = $self->getMaxCol20($j);

      $self->{res}[$i][$j] = 'P' if ($maxR == $maxC &&
				     $maxC == $self->{matrix}[$i][$j] &&
				     abs($j/$self->{size}[2]-$i/$self->{size}[1]) < $p);
    }
  }

  return $self;
}



=head2 C<getMaxRow20>

This function...

=cut

sub getMaxRow20 {
  my ($self, $r) = @_;

  my $max = 0.01;
  my $found = 0;

  for (0..$self->{size}[2]-1) {
    if ($self->{matrix}[$r][$_] > $max) {
      if ($self->{matrix}[$r][$_] > $max + 10) {
	$max = $self->{matrix}[$r][$_];
	$found = 1;
      } else {
	$found = 0;
      }
    }
  }
  return $found?$max:-1;
}

=head2 C<getMaxCol20>

This function...

=cut

sub getMaxCol20 {
  my ($self, $c) = @_;

  my $max = 0.01;
  my $found = 0;

  for (0..$self->{size}[1]-1) {
    if ($self->{matrix}[$_][$c] > $max) {
      if ($self->{matrix}[$_][$c] > $max + 10) {
	$max = $self->{matrix}[$_][$c];
	$found = 1;
      } else {
	$found = 0;
      }
    }
  }
  return $found?$max:-1;
}

=head2 C<prob>

This function...

=cut

sub prob {
  my ($self,$l,$r) = @_;

  my ($Size1,$Size2) = ($self->{size}[1],
			$self->{size}[2]);

  my $ret = 0;

  my $ll = NAT::Client::query_dictionary({Direction => "~>"}, $self->{crp}, $l);
  my $rr = NAT::Client::query_dictionary({Direction => "<~"}, $self->{crp}, $r);

  if ($ll && exists($ll->[1]{$r})) {
    $ret += 100*$ll->[1]{$r}
  }

  if ($rr && exists($rr->[1]{$l})) {
    $ret += 100*$rr->[1]{$l}
  }

  $ret /= 2;

  $ret;
}

=head2 C<enhanceDiagonal>

This function...

=cut

sub enhanceDiagonal {
  my ($self) = @_;

  for my $i (0..$self->{size}[1]-1) {
    for my $j (0..$self->{size}[2]-1) {
      $self->{matrix}[$i][$j] ||= 0;
      $self->{matrix}[$i][$j] *= (1-abs($i/$self->{size}[1] - $j/$self->{size}[2]))
    }
  }
}


1;
__END__

=head1 SEE ALSO

See perl(1) and NATools documentation.

=head1 AUTHOR

Alberto Manuel Brandao Simoes, E<lt>albie@alfarrabio.di.uminho.ptE<gt>

=head1 COPYRIGHT AND LICENSE

Copyright 2002-2005 by NATURA Project
http://natura.di.uminho.pt

This library is free software; you can redistribute it and/or modify
it under the GNU General Public License 2, which you should find on
parent directory. Distribution of this module should be done including
all NATools package, with respective copyright notice.

=cut

